from datetime import datetime
from typing import List, Optional, Sequence, Tuple
from itertools import groupby
from django.utils.translation import pgettext_lazy

from rest_framework_simplejwt.exceptions import InvalidToken
from rest_framework_simplejwt.tokens import Token as TokenStruct
from rest_framework_simplejwt.utils import datetime_from_epoch

# FIXME: This should not be here!
from wcd_device_recognizer.utils import model_bulk_get_or_create

from ..utils import get_token_jti
from ..models import Token


__all__ = 'check_expired', 'expire', 'register',

TokenPair = Tuple[TokenStruct, Optional[TokenStruct]]


def check_expired(token: TokenStruct, silent: bool = True) -> bool:
    exists = Token.objects.filter(jti=get_token_jti(token)).expired().exists()

    if not silent:
        raise InvalidToken(
            pgettext_lazy('wcd_jwt', 'Token is not longer valid.'),
            code='token_expired',
        )

    return exists


def expire(tokens: Sequence[TokenStruct], affect_tree: bool = False):
    tree = (
        Token.objects
        .filter(jti__in=map(get_token_jti, tokens))
        .values_list('id', flat=True)
    )

    if affect_tree:
        tree = Token.objects.collect_tree(ids=tree)

    if len(tree) == 0:
        return []

    Token.objects.filter(id__in=tree).set_expired()

    return tree


def simplify_token(token: TokenStruct) -> Tuple[str, str, datetime]:
    return (
        get_token_jti(token), str(token),
        datetime_from_epoch(token.payload['exp']),
    )


def _count_nested_list(nested: List[list], depth: int) -> int:
    # Recursion fix.
    if depth == 0:
        return 0

    depth -= 1

    return len(nested) + sum((
        _count_nested_list(inside, depth)
        for inside in nested
    ), 0)


def register(pairs: Sequence[TokenPair]) -> List[TokenPair]:
    resolved = [
        (
            simplify_token(child),
            simplify_token(parent) if parent is not None else None,
        )
        for child, parent in pairs
    ]
    parents = {}
    ids = {}

    # Collecting strange parents tree to find out how many parents to
    # resolve each of children has.
    for child, parent in resolved:
        parents[child] = parents.get(child, [])

        if parent is not None:
            parents[parent] = parents.get(parent, [])
            parents[child].append(parents[parent])

    total = len(parents)
    weight_sorted = sorted([
        (_count_nested_list(nested, total), i)
        for i, nested in parents.items()
    ])
    chunks = groupby(weight_sorted, key=lambda x: x[0])
    parents_map = dict(resolved)
    result = {}

    for _, items in chunks:
        items = [v for _, v in items]

        tokens = model_bulk_get_or_create(Token, [
            (
                {
                    'jti': value[0], 'token': value[1],
                },
                {
                    'parent_id': ids.get(parents_map.get(value)),
                    'expired_at': value[2]
                },
            )
            for value in items
        ])
        result.update(zip(items, tokens))
        ids.update(zip(items, (x.id for x in tokens)))

    return [
        (result[child], result[parent] if parent is not None else None)
        for child, parent in resolved
    ]



from dataclasses import dataclass
from px_settings.contrib.django import settings as s


__all__ = 'Settings', 'settings',


@s('WCD_JWT_REGISTRY')
@dataclass
class Settings:
    """
    Example:

    ```python
    WCD_JWT_REGISTRY = {
        'TOKEN_EXPIRE_SERIALIZER': 'wcd_jwt.contrib.registry.serializers.TokenExpireSerializer',
    }
    ```
    """
    TOKEN_EXPIRE_SERIALIZER: str = 'wcd_jwt.contrib.registry.serializers.TokenExpireSerializer'


settings = Settings()



from django.contrib import admin

from .models import Token, TokenUserConnection


@admin.register(Token)
class TokenAdmin(admin.ModelAdmin):
    list_display = 'id', 'jti', 'parent_id', 'expired_at', 'created_at',
    date_hierarchy = 'created_at'
    autocomplete_fields = 'parent',
    search_fields = 'id', 'jti', 'token',


@admin.register(TokenUserConnection)
class TokenUserConnectionAdmin(admin.ModelAdmin):
    list_display = 'id', 'user', 'token', 'created_at',
    date_hierarchy = 'created_at'
    list_select_related = 'user', 'token',
    autocomplete_fields = 'user', 'token',
    search_fields = 'id', 'user__username', 'token__token', 'token__jti',



from django.contrib.auth import authenticate
from rest_framework import exceptions, serializers

from rest_framework_simplejwt.settings import api_settings
from rest_framework_simplejwt.tokens import RefreshToken, UntypedToken
from rest_framework_simplejwt.serializers import (
    TokenObtainSerializer as _TokenObtainSerializer

)

from .conf import settings


class TokenObtainSerializer(_TokenObtainSerializer):
    def validate(self, attrs):
        authenticate_kwargs = {
            self.username_field: attrs[self.username_field],
            "password": attrs["password"],
        }
        try:
            authenticate_kwargs["request"] = self.context["request"]
        except KeyError:
            pass

        self.user = authenticate(**authenticate_kwargs)

        if not api_settings.USER_AUTHENTICATION_RULE(self.user):
            raise exceptions.AuthenticationFailed(
                self.error_messages["no_active_account"],
                "no_active_account",
            )

        return {'user': self.user}

    def commit(self):
        return {}


class TokenObtainPairSerializer(TokenObtainSerializer):
    token_class = RefreshToken

    def validate(self, attrs):
        data = super().validate(attrs)
        data['refresh'] = self.get_token(data['user'])
        data['access'] = data['refresh'].access_token

        return data

    def commit(self):
        v = self.validated_data

        return {'refresh': str(v['refresh']), 'access': str(v['access'])}


class TokenRefreshSerializer(serializers.Serializer):
    refresh = serializers.CharField()
    access = serializers.CharField(read_only=True)
    token_class = RefreshToken

    def validate(self, attrs):
        data = super().validate(attrs)
        data['refresh'] = self.token_class(attrs['refresh'])
        data['access'] = data['refresh'].access_token

        return data

    def commit(self):
        refresh = self.validated_data['refresh']
        data = {'access': str(self.validated_data['access'])}

        if settings.ROTATE_REFRESH_TOKENS:
            # TODO: Rework it to make some sort of token `.clone()`.
            updated = self.token_class(str(refresh))
            updated.set_jti()
            updated.set_exp()
            updated.set_iat()

            data["refresh"] = str(updated)

        return data


class TokenVerifySerializer(serializers.Serializer):
    token = serializers.CharField()

    def validate(self, attrs):
        # Will raise if something wrong with token.
        UntypedToken(attrs["token"])

        return attrs

    def commit(self):
        return {}



from rest_framework_simplejwt.authentication import JWTAuthentication as Base

from .services import token_manager


class JWTAuthentication(Base):
    def get_validated_token(self, raw_token):
        token = super().get_validated_token(raw_token)
        token_manager.check_expired(token, silent=False)

        return token



from django.urls import path
from django.views.decorators.csrf import csrf_exempt

from .conf import settings

from wcd_jwt.views import TokenViewBase


__all__ = (
    'TokenExpireView',
    'token_expire_view',
    'make_urlpatterns',
)


class TokenExpireView(TokenViewBase):
    """
    Expires any provided token.
    """
    _serializer_class = settings.TOKEN_EXPIRE_SERIALIZER


token_expire_view = TokenExpireView.as_view()


def make_urlpatterns(
    expire_view=token_expire_view,
):
    return [
        path('expire/', csrf_exempt(expire_view), name='refresh'),
    ]



from django.dispatch import receiver
from wcd_jwt.signals import token_refreshed, token_obtained
from django.contrib.auth import get_user_model

from .services import token_manager, user_connector


def _flat(pairs):
    return [y for x in pairs for y in x]


@receiver(token_obtained)
def register_token_on_obtain(sender, serializer, result, **kwargs):
    # TODO: Make this in a separate thread or even in celery.
    v = serializer.validated_data

    pairs = token_manager.register((
        (v['access'], v['refresh']),
    ))

    # Connecting users if possible.
    if 'user' in serializer.validated_data:
        user_connector.connect(
            serializer.validated_data['user'], _flat(pairs)
        )


@receiver(token_refreshed)
def register_token_on_refresh(sender, serializer, result, **kwargs):
    # TODO: Make this in a separate thread or even in celery.
    v = serializer.validated_data
    token_class = serializer.token_class
    base = v['refresh']
    pairs = [(v['access'], base)]

    if 'refresh' in result:
        refresh = token_class(result['refresh'])
        pairs = (
            # Wer'e replacing access token parent with newly generated refresh.
            # It's because we want to form a straight relation between
            # responded refresh->access tokens.
            (v['access'], refresh),
            (refresh, base),
        )

    pairs = token_manager.register(pairs)

    # Searching for user connections of the previous token(s).
    # To connect them. If there is more than one user - something
    # wrong happened and no connection will be established.
    users = list(
        get_user_model().objects
        .filter(token_connections__token__in=_flat(pairs))
    )
    if len(users) == 1:
        user_connector.connect(users[0], _flat(pairs))



from rest_framework_simplejwt.authentication import JWTAuthentication as Base
# TODO: Remove dependency over `api_settings`.
from rest_framework_simplejwt.settings import api_settings

from .services import token_resolver
from .signals import user_token_authenticated


class JWTAuthentication(Base):
    def get_validated_token(self, raw_token):
        return token_resolver.resolve_valid(
            raw_token, from_classes=api_settings.AUTH_TOKEN_CLASSES,
        )

    def authenticate(self, request):
        user, token = super().authenticate(request)
        user_token_authenticated.send(
            self.__class__, user=user, token=token, request=request
        )

        return user, token



from typing import Sequence, Type, Optional
from django.utils.translation import pgettext_lazy

from rest_framework_simplejwt.exceptions import InvalidToken, TokenError

from ..conf import settings
from ..utils import cached_import_string


def get_all_token_types():
    return [cached_import_string(x) for x in settings.TOKEN_TYPES]


def resolve_valid(
    raw_token: str,
    from_classes: Optional[Sequence[Type]] = None,
    silent: bool = False,
):
    """
    Validates an encoded JSON web token and returns a validated token
    wrapper object.
    """
    messages = []
    if from_classes is None:
        from_classes = get_all_token_types()

    for AuthToken in from_classes:
        try:
            return AuthToken(raw_token)
        except TokenError as e:
            messages.append({
                'token_class': AuthToken.__name__,
                'token_type': AuthToken.token_type,
                'message': e.args[0],
            })

    if silent:
        return None

    raise InvalidToken({
        'detail': pgettext_lazy(
            'wcd_jwt', 'Given token not valid for any token type'
        ),
        'messages': messages,
    })



from rest_framework_simplejwt.tokens import Token as TokenStruct
from rest_framework_simplejwt.settings import api_settings


def get_token_jti(token: TokenStruct) -> str:
    return token.payload[api_settings.JTI_CLAIM]



from typing import List, Sequence

# FIXME: This should not be here!
from wcd_device_recognizer.utils import model_bulk_get_or_create

from ..models import TokenUserConnection, Token


def connect(user, tokens: Sequence[Token]) -> List[TokenUserConnection]:
    uniques = {(user.pk, token.pk) for token in tokens}

    return model_bulk_get_or_create(TokenUserConnection, [
        ({'user_id': uid, 'token_id': tid}, {})
        for uid, tid in uniques
    ])
